'''# 网易云音乐 APIs'''
import base64
from logging import getLogger
from time import time
from requests.models import Response
from .. import GetCurrentSession,SetCurrentSession
from .. import logger
from ..utils import Crypto,ncm_client_key
import json,urllib.parse

class LoginRequiredException(Exception):pass
class LoginFailedException(Exception):pass
# region Base models export
LOGIN_REQUIRED = LoginRequiredException('Function needs you to be logged in')

def parse(url):return urllib.parse.urlparse(url)

def LoginRequiredApi(func):
    '''This API needs the user's token'''
    def wrapper(*a,**k):
        if not GetCurrentSession().login_info['success']:raise LOGIN_REQUIRED
        return func(*a,**k)
    return wrapper

def UserIDBasedApi(func):
    '''This API has UID as its first argument'''
    def wrapper(user_id=0,*a,**k):
        if user_id == 0 and GetCurrentSession().login_info['success']:
            user_id = GetCurrentSession().login_info['content']['account']['id']
        elif user_id == 0:
            raise LOGIN_REQUIRED
        return func(user_id,*a,**k)
    return wrapper

def BaseWrapper(requestFunc):
    '''Base wrapper for crypto requests
    '''
    def apiWrapper(apiFunc):
        '''apiFunc -> rtype : (url,plain,[method])'''
        def wrapper(*a,**k):
            ret       = apiFunc(*a,**k)
            url,plain = ret[:2]
            method    = ret[-1] if ret[-1] in ['POST','GET'] else 'POST'        
            rsp = requestFunc(url,plain,method)                        
            try:
                return json.loads(rsp.text) if isinstance(rsp,Response) else json.loads(rsp)
            except json.JSONDecodeError:
                return rsp
        return wrapper
    return apiWrapper

@BaseWrapper
def WeapiCryptoRequest(url,plain,method):
    '''This apis utilize `/weapi/` calls,seen in web & pc clients'''
    payload = json.dumps({**plain,'csrf_token':GetCurrentSession().csrf_token})
    return GetCurrentSession().request(method,
        url.replace('/api/','/weapi/'),
        params={'csrf_token':GetCurrentSession().csrf_token},
        data={**Crypto.WeapiCrypto(payload)},
    )
# region Port of `Binaryify/NeteaseCloudMusicApi`
@BaseWrapper
def LapiCryptoRequest(url,plain,method):
    '''This api proxies over /api/linux/forward,seen in Desktop Linux based clients'''
    payload = {
        'method':method,
        'url':GetCurrentSession().HOST + url,
        'params':plain
    }
    payload = json.dumps(payload)
    return GetCurrentSession().request(method,
        '/api/linux/forward',
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
        },
        data={**Crypto.LinuxCrypto(payload)}
    )

@BaseWrapper
def EapiCryptoRequest(url,plain,method):
    '''This API utilize '/api/' or '/eapi/',seen in mobile clients'''
    cookies = {
        'appver': '7.2.24', # the version used for decompilation
        'buildver':'whatever_the_latest_is', # ...these don't matter   
        'channel':'persumably_offical',
        'deviceId': 'pyncm_but_who_cares',   # ...
        'mobilename' : 'Pixel2XL',           # ...as long as these matches our UA
        'os': 'android',
        'osver':'10.0',                      # ...
        'requestId':f'{int(time() * 1000)}_0233',
        'resolution': '2712x1440',
        'versioncode': '240',
        '__csrf':GetCurrentSession().csrf_token,
        **GetCurrentSession().cookies
    }
    payload = {
        **plain,    
        'header':json.dumps(cookies)
    }
    request = GetCurrentSession().request(method,
        url,
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':'NeteaseMusic/7.2.24.1597753235(7002024);Dalvik/2.1.0 (Linux; U; Android 11; Pixel 2 XL Build/RP1A.200720.009)',
            'Referer':None
        },
        cookies=cookies,
        data={**Crypto.EapiCrypto(parse(url).path.replace('/eapi/','/api/'),json.dumps(payload))}
    )
    content = request.content
    try:
        content = Crypto.EapiDecrypt(content).decode()
    except Exception as e:
        getLogger('ncm').warn('Falling back to plaintext response:%s' % e)
    return content
# endregion
# endregion
from . import miniprograms,album,cloud,cloudsearch,login,playlist,track,user,video


'''歌曲 - Track APIs'''
from . import WeapiCryptoRequest,EapiCryptoRequest,LoginRequiredApi
from . import logger
import json
@WeapiCryptoRequest
def GetTrackDetail(song_ids : list):
    '''网页端 - 获取歌曲详情

    Args:
        song_ids (list): 歌曲 ID

    Returns:
        dict
    '''
    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/weapi/v3/song/detail',{"c":json.dumps([{'id': str(id)} for id in ids])}

@WeapiCryptoRequest
def GetTrackAudio(song_ids:list, quality='lossless',encodeType='aac'):
    '''网页端 - 获取歌曲音频详情（文件URL、MD5...）

    Args:
        song_ids (list): 歌曲 ID
        quality (str, optional): standard / high / lossless . Defaults to 'lossless'.
        encodeType (str, optional): 歌曲编码 . Defaults to 'aac'.

    Returns:
        dict
    '''
    if not quality in ['standard', 'higher', 'lossless']:
        logger.warn('Invalid quality config `%s`,falling back to `standard`')
        quality = 'standard'
    # Quality check
    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/weapi/song/enhance/player/url/v1',{"ids":ids,"level":quality,"encodeType":encodeType}


@WeapiCryptoRequest
def GetTrackLyrics(song_id : str):
    '''网页端 - 获取歌曲歌词

    Args:
        song_id (str): 歌曲ID

    Returns:
        dict
    '''
    return '/weapi/song/lyric',{"id": str(song_id), "lv": -1, "tv": -1}


@WeapiCryptoRequest
def GetTrackComments(song_id, offset=0, limit=20,beforeTime=0):
    '''网页端 - 获取歌曲评论

    Args:
        album_id (str): 歌曲ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/v1/resource/comments/R_SO_4_%s' % song_id,{"rid":str(song_id),"offset":str(offset),"total":"true","limit":str(limit),"beforeTime":str(beforeTime * 1000)}


@EapiCryptoRequest
@LoginRequiredApi
def SetLikeTrack(trackId,like=True,userid=0,e_r=True):
    '''PC端 - 收藏歌曲到 `我喜欢的音乐` - WIP  - WIP - protected via 网易易盾

    TODO: defeat `watchman.js` obfuscation
    
    Args:
        trackId (int): 歌曲 ID
        like (bool, optional): 收藏或取消收藏. Defaults to True.
        userid (int, optional): 暂存. Defaults to 0.

    Returns:
        dict
    '''
    return '/api/song/like',{"trackId":str(trackId),"userid":str(userid),"like":str(like).lower(),"e_r":str(e_r).lower()}    


'''Implementation of some of NE's crypto functions'''
import base64
from Crypto.Cipher import AES
from Crypto.Random import random
from hashlib import md5


class RSAPublicKey():
    '''
        RSA Publickey object,where

            n   :   modulus
            e   :   exponet
        All values are stored as integers
    '''
    def __init__(self, n, e):
        self.n = int(n, 16)
        self.e = int(e, 16)

# region Constant values
weapi_aes_key    = "0CoJUm6Qyw8W8jud" # cbc
weapi_aes_iv     = "0102030405060708" # cbc
weapi_rsa_pubkey = RSAPublicKey(
    "00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7",
    "10001" # signle-block,no padding
)
weapi_rsa_default_crypto = (
    'mos9527ItoooItop', 
    # plaintext (aes_key_2)
    '01a1c399271006da676da55763419f10f0e589515c49530b33418eec82202fc42dae0cd3aa4a2b7bdc3dafa7c6a918e405f3cdbc5d0349ef86913fc2dbe8764ed782e202e7828b547e85f6ae28b8b120bcf5fd3777a55731521612dcaff9813246a42876303b0f2307c9f264671ddc87159ff162e689fdfae5acb3af10250754'
    # (key->rsa_pubkey) ciphertext (encSecKey)
)   # signle-block,no padding
linuxapi_aes_key = "rFgB&h#%2?^eDg:Q" # ecb
eapi_digest_salt = "nobody%(url)suse%(text)smd5forencrypt"
eapi_data_salt   = "%(url)s-36cd479b6b5-%(text)s-36cd479b6b5-%(digest)s"
eapi_aes_key     = "e82ckenh8dichen8" # ecb
ncm_client_key   = "9ca17ae2e6ffcda170e2e6eed6c16ded96bf85b57498868ab2c45e939b8eaaf842a7ee85b0f47082eabf9bf02af0feaec3b92ab39fa889ed5d868682aab45a879f9ab6d84f9aaf9783f06bbc92a5afe8678391ee9e"
ac_token_1       = ("0b0cdd23ed1144a0b78de049edc09824","VMhXA/Eedu9FBARERUN/DadrGuFdDoXR")
ac_C_t           = "0eYIS6J6qRFUBFRAUVdvZazbsg2mWj6E"
randstr_charset  = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
# endregion

def PadWithRemainder(data,blocksize=AES.block_size):
    '''Pads with the remaing length of byte thats needs until our length is a intergal multipule of blocksize'''
    return data + (blocksize - len(data) % blocksize) * chr(blocksize - len(data) % blocksize)

def UnpadRemainder(data):
    '''The reverse funtion of `PadWithRemainder`'''
    pad = data[-1]
    if not pad in range(0,AES.block_size):return data # data isn't padded
    return data[:-pad]

class Crypto():
    '''The cryptography toolkit'''
    # region Base cryptograhpy methods
    # region Utility
    @staticmethod
    def RandomString(len, chars=randstr_charset):
        '''Generates random string of `len` chars within a selected number of chars'''
        return ''.join([random.choice(chars) for i in range(0, len)])
    @staticmethod
    def HexDigest(data : bytearray):
        '''Digests a `bytearray` to a hex string'''
        return ''.join([hex(d)[2:].zfill(2) for d in data])
    @staticmethod
    def HexCompose(hexstr : str):
        '''Composes a hex string back to a `bytearray`'''
        if len(hexstr) % 2:raise Exception('Hex-string length should be a even number')
        return bytearray([int(hexstr[i:i+2],16) for i in range(0,len(hexstr),2)])
    @staticmethod
    def HashDigest(text):
        '''Hexdecimal MD5 hash generator'''
        HASH = md5(text.encode('utf-8'))
        return HASH.hexdigest()
    # endregion
    # region Cryptos
    @staticmethod
    def AESEncrypt(
        data:str, key:str, mode=AES.MODE_ECB , iv='',padder=PadWithRemainder        
    ):
        '''Basic AES encipher function'''
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            encryptor = AES.new(key.encode(), mode)
        else:    
            encryptor = AES.new(key.encode(), mode, iv.encode())
        encrypt_aes = encryptor.encrypt(str.encode(padder(data)))
        return encrypt_aes

    @staticmethod
    def AESDecrypt(
        data:str,key:str,mode=AES.MODE_ECB,iv='',unpadder=UnpadRemainder
    ):
        '''Basic AES decipher funtion'''
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            decryptor = AES.new(key.encode(), mode)
        else:    
            decryptor = AES.new(key.encode(), mode, iv.encode())
        decrypted_aes = decryptor.decrypt(data)
        return unpadder(decrypted_aes)

    @staticmethod
    def RSAEncrypt(data:str, pubkey: RSAPublicKey, reverse=True):
        '''Signle-block textbook RSA encrpytion (c ≡ n ^ e % N),encodes text to hexstring first'''
        def toInt(s): return int(s.encode('utf-8').hex(), 16)
        n = data # do not modify in place
        if reverse:n = reversed(n)
        n, e, N = toInt(''.join(n)), pubkey.e, pubkey.n
        return Crypto.HexCompose(hex(n ** e % N)[2:].zfill(256))
    # endregion
    # endregion
    
    # region Netease crypto
    '''source:https://s3.music.126.net/web/s/core_53f8fab3b18ff334dd7af53eec462c5e.js'''
    @staticmethod
    def WeapiCrypto(params, crypto=weapi_rsa_default_crypto):
        '''Used in web & PC client'''
        aes_key2,encSecKey = crypto
        params = str(params)
        # 1st go,encrypt the text with aes_key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=weapi_aes_key, iv=weapi_aes_iv,mode=AES.MODE_CBC)), encoding='utf-8')
        # 2nd go,encrypt the ENCRYPTED text again,with the 2nd key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=aes_key2, iv=weapi_aes_iv,mode=AES.MODE_CBC)), encoding='utf-8')
        # 3rd go,generate RSA encrypted encSecKey
        if not encSecKey:
            # try not to generate this on-the-fly as it's quite slow
            encSecKey = Crypto.HexDigest(Crypto.RSAEncrypt(aes_key2, weapi_rsa_pubkey))
        return {
            'params': params,
            'encSecKey': encSecKey
        }

    '''source:
        - decompilation of `libpoison.so` : https://juejin.im/user/2383396938455821
        - Binaryify/NeteaseCloudMusicApi  : https://github.com/Binaryify/NeteaseCloudMusicApi/blob/master/util/crypto.js'''
    @staticmethod
    def LinuxCrypto(params):
        '''Used in desktop linux clients'''
        params = str(params)
        # Single-pass AES-128 ECB crypto
        return {
            'eparams':Crypto.HexDigest(Crypto.AESEncrypt(params,key=linuxapi_aes_key,mode=AES.MODE_ECB))
        }
    
    @staticmethod
    def EapiCrypto(url,params):
        '''Used in mobile clients'''
        url,params = str(url),str(params)
        digest = Crypto.HashDigest(eapi_digest_salt % {'url':url,'text':params})
        params = eapi_data_salt % ({'url':url,'text':params,'digest':digest})
        return {
            'params':Crypto.HexDigest(Crypto.AESEncrypt(params,key=eapi_aes_key,mode=AES.MODE_ECB))
        }

    @staticmethod
    def EapiDecrypt(cipher):
        '''Used in mobile clients'''
        cipher = bytearray(cipher) if isinstance(cipher,str) else cipher
        return Crypto.AESDecrypt(cipher,eapi_aes_key,mode=AES.MODE_ECB) if cipher else cipher
    # endregion


'''我的音乐云盘 - Cloud APIs'''
from . import WeapiCryptoRequest,LoginRequiredApi

@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveInfo(limit=30,offset=0):
    '''PC端 - 获取个人云盘内容

    Args:
        limit (int, optional): 单次获取量. Defaults to 30.
        offset (int, optional): 获取偏移数. Defaults to 0.

    Returns:
        dict
    '''
    return '/weapi/v1/cloud/get',{limit:str(limit),offset:str(offset)}

@WeapiCryptoRequest
@LoginRequiredApi
def GetCloudDriveItemInfo(song_ids:list):
    '''PC端 - 获取个人云盘项目详情

    Args:
        song_ids (list): 云盘项目 ID

    Returns:
        dict
    '''
    ids = song_ids if isinstance(song_ids,list) else [song_ids]
    return '/weapi/v1/cloud/get/byids',{'songIds':ids}


'''Pool based multithreaded downloader'''
import os,sys,time
from threading import Thread
from queue import Queue
from requests import Session

class PoolWorker(Thread):
    '''Base worker model.'''
    def __init__(self,queue,id=0):
        self.id = id
        self.task_queue = queue
        super().__init__()
        # This is needed for daemon flag to be set.
        self.daemon = True
        # Once set,the main thread won't stay to wait for its sub-threads to end

    def run(self):
        while True:            
            task,args = self.task_queue.get()
            try:
                task(args) if args else task()
            except Exception as e:
                sys.stderr.write(f'Worker exception:{e}\n')            
            self.task_queue.task_done()

class DownloadWorker(PoolWorker):
    '''Workers for downloading'''

    def init_status(self):
        self.status = {'id': self.id, 'status': None, 'url': None,
                       'path': None, 'length': None, 'xfered': 0}

    def __init__(self, session: Session,task_queue: Queue, id=0,  timeout=5,buffer_size=256):
        super().__init__(task_queue,id)
        self.session = session
        self.timeout = timeout
        self.buffer_size = buffer_size
        # Buffer size.Note it's in KB
        self.init_status()


    def __call__(self):
        '''Reports the status once called'''
        return self.status

    def run(self):
        while True:
            url,path = self.task_queue.get()
            # This will block the thread if no queue object is available,work start        
            self.init_status()
            # Reinitalize status
            self.status = {**self.status, 'url':url,'path':path}
            r = None
            try:
                r = self.session.get(url, stream=True, timeout=self.timeout)
                self.status['status'] = r.status_code
                # Send GET request and send stream flag to download data in a streamed fashion
            except Exception:
                self.status['status'] = 0x198
                # Time-out - failed to get the requested resource (code 408)
            if not self.status['status'] == 0xC8:
                self.task_queue.task_done()
                continue
            # Stops this iteration if server doesn't response 200 (0xC8)
            length = int(r.headers['content-length'])
            if (length):
                self.status['length'] = length
            # Sets content-length if the server ever sends one
            p = os.path.split(path)
            if p[0]:
                if not os.path.exists(p[0]):os.makedirs(p[0])
            # Creates directory tree if dosen't exsist
            try:
                with open(path, 'wb') as f:
                    chunk_size = self.buffer_size * 1024
                    # Buffer of 128 KB
                    for chunk in r.iter_content(chunk_size):
                        # Iterate content with buffer size of which
                        f.write(chunk)
                        self.status['xfered'] += len(chunk)
                        # Writes to file and updates infomation
                self.status['status'] = 0xFF
                # sets flag to 0xFF to flag completion since 256 isnt in the HTTP Standard 
            except Exception as e:
                # Uncaught excpetion
                sys.stderr.write(e)
            self.task_queue.task_done()
            # Marks that one task is completed.Note that it doesn't specifiy which task,work end
class Downloader():
    '''Threadpool a-like downloader

    Args :
    
            session   :   request.Session()
            pool_size :   downloader count,specifies how many cocurrent tasks can be processed at once
            timeout   :   the time before raising TimeoutException
            append    :   append a new download task
            wait      :   wait util all tasks are finished
            report    :   geneartes a pretty report
    '''

    def __init__(self,session: Session = None, worker=DownloadWorker, pool_size=1, timeout=5,buffer_size=256):
        if not session:session = Session()
        self.pool_size = pool_size
        self.task_queue = Queue()
        def get_worker(i):
            if worker == DownloadWorker:
                return worker(session,self.task_queue,id=i, timeout=timeout,buffer_size=buffer_size) 
            elif worker == PoolWorker:
                return worker(self.task_queue,id=i) 
            else:
                raise NotImplementedError
        self.workers = [get_worker(i) for i in range(0, pool_size) ]
        # Generate workers
        for worker in self.workers:
            worker.start()
        # and start them

    def reports(self):
        '''Generates reports.'''
        for worker in self.workers:
            if not type(worker) == DownloadWorker:
                raise NotImplementedError(type(worker))
            status,id,xfered,length = worker()['status'],worker()['id'],worker()['xfered'],worker()['length']            
            yield status,id,xfered,length

        
    def wait(self, *args, func=None, do_when_done=True):
        '''Equvilant to .task_queue.join(),but tasks during wait time is possible.
                do_when_done    :   Specifies whether exec the fucntion when loop ends or not
            Waits for all tasks are finished
        '''
        def do_func():
            if args and func:
                func(*args)
            elif func:
                func()

        while self.task_queue.unfinished_tasks != 0:
            do_func()
        if do_when_done:
            do_func()

    def append(self, url, path):
        '''Appends a new download task into queue

                url     : the file url
                path    : the path of the destination file
        '''
        self.task_queue.put((url,path))
